{% extends "!search.html" %}
{% block extrahead %}
    {{ super() }}
    <meta name="robots" content="noindex">
    <script type="text/javascript" src="{{ pathto('_static/glossary_search.js', resource=True) }}"></script>
    <meta property="og:title" content="Search — Python {{ version }} documentation">
    <meta property="og:type" content="website">
    <meta property="og:url" content="https://docs.python.org/3/search.html">
    <meta property="og:site_name" content="Python documentation">
    <meta property="og:description" content="Search the Python documentation.">
    <meta property="og:image" content="https://docs.python.org/3/_static/og-image.png">
    <meta name="description" content="Search the Python documentation.">
    <meta property="og:image:width" content="200">
    <meta property="og:image:height" content="200">
    <meta name="theme-color" content="#3776ab">
{% endblock %}
{% block searchresults %}
<div id="search-results">
  {# For glossary_search.js #}
  <div style="display: none;" class="admonition seealso" id="glossary-result">
    <p class="topic-title">
    <a id="glossary-title" href="#"></a>
    </p>
  <div id="glossary-body"></div>
</div>
</div>
{% endblock %}
